"""
Functions for planet imagery and interacting with DE Africa dataset.
"""

import os
import warnings
from datetime import datetime

import datacube
import numpy as np
from ipyleaflet import (
    FullScreenControl,
    GeoData,
    LayersControl,
    LegendControl,
    Map,
    TileLayer,
    basemap_to_tiles,
    basemaps,
)
from localtileserver import TileClient, get_leaflet_tile_layer
from odc.geo.xr import write_cog

from deafrica_tools.bandindices import calculate_indices
from deafrica_tools.waterbodies import get_waterbodies

# Turn off all warnings.
warnings.filterwarnings("ignore")
warnings.simplefilter("ignore")

os.environ["LOCALTILESERVER_CLIENT_PREFIX"] = (
    f"{os.environ['JUPYTERHUB_SERVICE_PREFIX']}/proxy/{{port}}"
)


def get_last_calendar_month() -> tuple[int, int]:
    """
    Get the year and month of the previous calendar month.
    """
    # Get the current date
    today = datetime.today()

    current_year = today.year
    current_month = today.month

    # Extract the year and month of the previous month
    if current_month == 1:
        year = current_year - 1
        month = 12
    else:
        year = current_year
        month = current_month - 1

    return year, month


# Load map to display
def loadplanet(lon_range: tuple, lat_range: tuple, threshold_nvdi: float, threshold_bui: float):
    """
    Loading of planet imagery and S2 R
    """
    # Connect to the datacube
    dc = datacube.Datacube(app="planet")

    # Create a bounding box from study area coordinates
    bbox = (lon_range[0], lat_range[0], lon_range[1], lat_range[1])

    # Select all water bodies located within the bounding box
    polygons = get_waterbodies(bbox, crs="EPSG:4326")
    # Identify columns with non-serializable data in the water polygons
    non_serializable = polygons.select_dtypes(include=["datetime64[ns]", "datetime64"]).columns
    # Drop columns
    polygons = polygons.drop(columns=non_serializable)

    # Get the year and month of the previous calendar month.
    year, month = get_last_calendar_month()

    # Load the S2 Annual GeoMAD
    ds = dc.load(
        product="gm_s2_annual",
        measurements=["red", "green", "blue", "nir", "swir_1"],
        y=lat_range,
        x=lon_range,
        resolution=(-10, 10),
        time=(f"{year:04d}"),
    )
    # Check if there was any current data in the previous load
    if len(ds) == 0:
        # if  not then load with the manual date provider
        ds = dc.load(
            product="gm_s2_annual",
            measurements=["red", "green", "blue", "nir", "swir_1"],
            y=lat_range,
            x=lon_range,
            resolution=(-10, 10),
            time=("2022", f"{year:04d}"),
        )

    # select the last best image
    ds = ds.isel(time=-1)

    # calculate NDVI and BUI
    ds = calculate_indices(ds, index=["NDVI", "BUI"], satellite_mission="s2")

    # Select NDVI value greater than threshold_nvdi value
    ds_ndvi = ds.where(ds.NDVI >= threshold_nvdi, np.nan).NDVI

    # Save the ndvi raster file
    write_cog(ds_ndvi, fname="ndvi.tif", overwrite=True)

    # Select BUI greater than 0
    ds_bui = ds.where(ds.BUI >= threshold_bui, np.nan).BUI

    # Save the bui raster file
    write_cog(ds_bui, fname="bui.tif", overwrite=True)

    # Defining the planet api
    planet_ = (
        f"https://api.digitalearth.africa/planet/tiles/basemaps/v1/planet-tiles/planet_medres_visual_{year:04d}-{month:02d}_mosaic/gmap/"
        + "{z}/{x}/{y}.png"
    )

    # Converting the api to tilelayer
    planet_tile = TileLayer(
        url=planet_,
        name=f"Planet NICFI-{year:04d}-{month:02d}",
        show_loading=True,
        attribution="Planet NICFI",
    )
    planet_tile.base = True

    # Identify columns with non-serializable data in the water polygons
    non_serializable = polygons.select_dtypes(include=["datetime64[ns]", "datetime64"]).columns
    # Drop columns
    polygons = polygons.drop(columns=non_serializable)
    # Converting the water polygons to GeoData
    geo_data = GeoData(
        geo_dataframe=polygons,
        style={
            "color": "black",
            "fillColor": "#3366cc",
            "opacity": 0.05,
            "weight": 1.9,
            "dashArray": "2",
            "fillOpacity": 0.6,
        },
        name="Water Body",
    )

    # Create a tile server from local raster file
    bui_client = TileClient("bui.tif")
    ndvi_client = TileClient("ndvi.tif")

    # Create ipyleaflet tile layer from that server
    tile_info = f"<a href=https://explorer.digitalearth.africa/products/gm_s2_annual> DE Africa Sentinel-2 Annual GeoMAD {ds.time.dt.year.values} </a>"
    bui = get_leaflet_tile_layer(
        bui_client,
        nodata=np.nan,
        name="Built up area",
        colormap="Reds",
        vmin=-1,
        vmax=1,
        attribution=tile_info,
    )
    ndvi = get_leaflet_tile_layer(
        ndvi_client, nodata=np.nan, name="Vegetation", colormap="Greens", vmin=0, vmax=1
    )

    # Openstreet basemap
    openstreet = basemap_to_tiles(basemaps.OpenStreetMap.Mapnik)
    openstreet.base = True

    # Create ipyleaflet map, add tile layer, and display
    map_con = Map(center=bui_client.center(), zoom=bui_client.default_zoom + 2)

    control = LayersControl(position="topright", collapsed=False)

    legend = LegendControl(
        {
            "Built-up Area": "#ff8533",
            "\n": "#fff",
            "Vegetation": "#34a832",
            " ": "#fff",
            "Water body": "#1d18c4",
        },
        title="Legend",
        position="bottomleft",
    )

    # add the created layer and control to the map
    map_con.add(bui)
    map_con.add(ndvi)
    map_con.add(geo_data)
    map_con.add(openstreet)
    map_con.add(planet_tile)
    map_con.add(FullScreenControl())
    map_con.add(control)
    map_con.add(legend)

    return map_con



"""
Loading OECD SWAC Africapolis data.
Last modified: December 2024
"""

import requests
import geopandas as gpd
from shapely.geometry import box
from owslib.wfs import WebFeatureService


WFS_ADDRESS = "https://geoserver.digitalearth.africa/geoserver/wfs"

def get_africapolis(
    bbox, crs="EPSG:4326", layer=None
) -> gpd.GeoDataFrame:
    """
    Retrieve Africapolis data from a GeoServer WFS for a given bounding box.
    Parameters
    ----------
    bbox : tuple or GeoDataFrame
        Bounding box as (xmin, ymin, xmax, ymax) in the specified CRS,
        or a GeoPandas object to extract bounds from.
    crs : str
        Coordinate Reference System for the bounding box. Defaults to EPSG:4326.
    layer : str
        Africapolis layer to load ('africapolis_2015' or 'africapolis_2020'). 
    drop_wms : bool
        Drop visualization-specific columns. Defaults to True.
    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame containing Africapolis data.
    """
    # Extract bounding box if a GeoDataFrame is provided
    if isinstance(bbox, gpd.GeoDataFrame) or isinstance(bbox, gpd.GeoSeries):
        crs = str(bbox.crs)
        bbox = bbox.total_bounds
    # Validate the layer choice (either 2015 or 2020)
    describe_url = (
        f"https://geoserver.digitalearth.africa/geoserver/wms?service=WMS&version=1.1.1&"
        f"request=DescribeLayer&layers=africapolis:{layer}&outputFormat=application/json"
    )
    try:
        describe_response = requests.get(describe_url).json()
        available_layers = [
            l["layerName"] for l in describe_response.get("layerDescriptions", [])
        ]
        # if layer not in available_layers:
        #     raise ValueError(f"Layer {layer} is not available. Available layers: {available_layers}")
    except Exception as e:
        raise RuntimeError(f"Error retrieving available layers: {e}")
    # Query the WFS for the specified layer and bbox
    try:
        wfs = WebFeatureService(url=WFS_ADDRESS, version="2.0.0")
        response = wfs.getfeature(
            typename=f"africapolis:{layer}",
            bbox=tuple(bbox) + (crs,),
            outputFormat="application/json"
        )
    except Exception as e:
        raise RuntimeError(f"Error querying WFS: {e}")
    # Load data into a GeoDataFrame
    africapolis_gdf = gpd.read_file(response)

    return africapolis_gdf


"""
Loading and processing DE Africa Water Bodies data.
Last modified: November 2023
"""

from datetime import datetime

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from matplotlib.patches import Patch
from owslib.etree import etree
from owslib.fes import PropertyIsEqualTo
from owslib.wfs import WebFeatureService

# URL for the DE Africa Water Bodies data on PROD Geoserver.
WFS_ADDRESS = "https://geoserver.digitalearth.africa/geoserver/wfs"
WFS_LAYER = "waterbodies:DEAfrica_Waterbodies"
API_ADDRESS = "https://api.digitalearth.africa/waterbodies/"


def get_waterbody(geohash: str) -> gpd.GeoDataFrame:
    """Gets a waterbody polygon and metadata by geohash.

    Parameters
    ----------
    geohash : str
        The geohash/uid for a waterbody in DE Africa Water Bodies.

    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame with the polygon.
    """

    wfs = WebFeatureService(url=WFS_ADDRESS, version="1.1.0")
    filter_ = PropertyIsEqualTo(propertyname="uid", literal=geohash)
    filterxml = etree.tostring(filter_.toXML()).decode("utf-8")
    response = wfs.getfeature(
        typename=WFS_LAYER,
        filter=filterxml,
        outputFormat="json",
    )
    wb_gpd = gpd.read_file(response)
    return wb_gpd


def get_waterbodies(bbox: tuple, crs: str = "EPSG:4326") -> gpd.GeoDataFrame:
    """Gets the polygons and metadata for multiple water bodies by bbox.

    Parameters
    ----------
    bbox : (xmin, ymin, xmax, ymax)
        Bounding box.
    crs : str
        Optional CRS for the bounding box.

    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame with the polygons and metadata.
    """

    wfs = WebFeatureService(url=WFS_ADDRESS, version="1.1.0")
    response = wfs.getfeature(
        typename=WFS_LAYER,
        bbox=tuple(bbox) + (crs,),
        outputFormat="json",
    )
    wb_gpd = gpd.read_file(response)
    return wb_gpd


def get_geohashes(bbox: tuple = None, crs: str = "EPSG:4326") -> list[str]:
    """Gets all waterbody geohashes.

    Parameters
    ----------
    bbox : (xmin, ymin, xmax, ymax)
        Optional bounding box.
    crs : str
        Optional CRS for the bounding box.

    Returns
    -------
    [str]
        A list of geohashes.
    """

    wfs = WebFeatureService(url=WFS_ADDRESS, version="1.1.0")
    if bbox is not None:
        bbox = bbox + (crs,)
    response = wfs.getfeature(
        typename=WFS_LAYER,
        propertyname="uid",
        outputFormat="json",
        bbox=bbox,
    )
    wb_gpd = gpd.read_file(response)
    return list(wb_gpd["uid"])


def get_time_series(
    geohash: str = None,
    waterbody: pd.Series = None,
    start_date: str = "1984-01-01",
    end_date: str = datetime.now().strftime("%Y-%m-%d"),
) -> pd.DataFrame:
    """Gets the time series for a waterbody. Specify either a GeoDataFrame row or a geohash.

    Parameters
    ----------
    geohash : str
        The geohash/uid for a waterbody in DE Africa Water Bodies.
    waterbody : pd.Series
        One row of a GeoDataFrame representing a waterbody.
    start_date : str
        Start date for the time range to filter the timeseries to.
    end_date : str
        End date for the time range to filter the timeseries to.
    Returns
    -------
    pd.DataFrame
        A time series for the waterbody.
    """
    if waterbody is not None and geohash is not None:
        raise ValueError("One of waterbody and geohash must be None")
    if waterbody is None and geohash is None:
        raise ValueError("One of waterbody and geohash must be specified")

    if geohash is not None:
        wb = get_waterbody(geohash)
        wb_id = wb.wb_id.item()
    else:
        wb_id = waterbody.wb_id.item()
    url = (
        API_ADDRESS
        + f"waterbody/{wb_id}/observations/csv?start_date={start_date}&end_date={end_date}"
    )
    wb_timeseries = pd.read_csv(url)
    # Tidy up the dataframe.
    wb_timeseries = wb_timeseries.set_index("date")
    wb_timeseries.index = pd.to_datetime(wb_timeseries.index)
    # Create a rolling median for the wet time series
    wb_timeseries["percent_wet_rolling_median"] = wb_timeseries["percent_wet"].rolling(3).median()

    return wb_timeseries


def display_time_series(wb_timeseries: pd.DataFrame = None) -> None:
    """Displays the timeseries as an interactive plot

    Parameters
    ----------
    wb_timeseries : pd.DataFrame
        A time series for the waterbody.

    Returns
    -------
    None
    """

    fig = go.Figure()

    # Add a scatter plot of invalid percentage measruements
    fig.add_trace(
        go.Scatter(
            x=wb_timeseries.index,
            y=wb_timeseries["percent_invalid"],
            mode="markers",
            marker=dict(color="red"),
            name="Invalid Percentage",
            hovertemplate="<b>%{x|%Y-%m-%d}</b><br>Invalid: %{y:.2f}%<extra></extra>",
            opacity=0.7,
        )
    )

    # Add a line for the median wet percentage
    fig.add_trace(
        go.Scatter(
            x=wb_timeseries.index,
            y=wb_timeseries["percent_wet_rolling_median"],
            mode="lines",
            line=dict(color="blue"),
            opacity=0.3,
            name="Wet Percentage - Rolling Median",
        )
    )

    # Add a scatter plot of the wet percentage measurements
    fig.add_trace(
        go.Scatter(
            x=wb_timeseries.index,
            y=wb_timeseries["percent_wet"],
            mode="markers",
            marker=dict(color="blue"),
            name="Wet Percentage",
            hovertemplate="<b>%{x|%Y-%m-%d}</b><br>Wet: %{y:.2f}%<extra></extra>",
            opacity=0.7,
        )
    )

    # Customize layout
    fig.update_layout(
        title="Wet surface area time series",
        xaxis_title="Date",
        yaxis_title="Percentage",
        yaxis_range=[-5, 105],
    )

    # Display the plot
    fig.show()


def plot_last_valid_obs(waterbodies: gpd.GeoDataFrame):
    """
    Plot the waterbodies in based on the value in the "last_valid_obs" column.
    Parameters
    ----------
    waterbodies: gpd.GeoDataFrame
        Waterbodies to plot

    Returns
    -------

    """
    assert "last_valid_obs" in waterbodies.columns

    # Assuming 'polygons' is a GeoDataFrame and you already have it loaded
    df = waterbodies.copy(deep=True)

    # List of conditions to color the polygons by
    conditions = [
        (df["last_valid_obs"] >= 0) & (df["last_valid_obs"] <= 10),
        (df["last_valid_obs"] > 10) & (df["last_valid_obs"] <= 20),
        (df["last_valid_obs"] > 20) & (df["last_valid_obs"] <= 30),
        (df["last_valid_obs"] > 30) & (df["last_valid_obs"] <= 40),
        (df["last_valid_obs"] > 40) & (df["last_valid_obs"] <= 50),
        (df["last_valid_obs"] > 50) & (df["last_valid_obs"] <= 60),
        (df["last_valid_obs"] > 60) & (df["last_valid_obs"] <= 70),
        (df["last_valid_obs"] > 70) & (df["last_valid_obs"] <= 80),
        (df["last_valid_obs"] > 80) & (df["last_valid_obs"] <= 90),
        (df["last_valid_obs"] > 90) & (df["last_valid_obs"] <= 100),
    ]

    # List of corresponding colors
    colors = [
        "#fde725",
        "#b5de2b",
        "#6ece58",
        "#35b779",
        "#1f9e89",
        "#26828e",
        "#31688e",
        "#3e4989",
        "#482878",
        "#440154",
    ]

    # Create a new column based on conditions
    df["color"] = np.select(conditions, colors, default="hatch")

    # Plot the polygons by color
    fig, ax = plt.subplots(figsize=(10, 10))

    # Separate colored polygons from those with hatch
    hatched = df[df["color"] == "hatch"]
    colored = df[df["color"] != "hatch"]

    # Plot the colored polygons
    colored.plot(ax=ax, color=colored["color"], edgecolor="none", linewidth=0.5)

    # Plot the hatched polygons
    hatched.plot(ax=ax, hatch="xx", facecolor="none", edgecolor="#23d9f1", linewidth=0.5)

    # Create the legend manually
    legend_elements = [
        Patch(color="#fde725", label="0 <= wet percentage <= 10"),
        Patch(color="#b5de2b", label="10 < wet percentage <= 20"),
        Patch(color="#6ece58", label="20 < wet percentage <= 30"),
        Patch(color="#35b779", label="30 < wet percentage <= 40"),
        Patch(color="#1f9e89", label="40 < wet percentage <= 50"),
        Patch(color="#26828e", label="50 < wet percentage <= 60"),
        Patch(color="#31688e", label="60 < wet percentage <= 70"),
        Patch(color="#3e4989", label="70 < wet percentage <= 80"),
        Patch(color="#482878", label="80 < wet percentage <= 90"),
        Patch(color="#440154", label="90 < wet percentage <= 100"),
        Patch(
            facecolor="none", hatch="xx", edgecolor="#23d9f1", label="No valid observation"
        ),  # Add hatch legend
    ]

    # Add the legend to the plot
    ax.legend(
        handles=legend_elements,
        title="Last Valid Observation",
        loc="upper left",
        bbox_to_anchor=(1, 1),
    )

    # Show the plot
    plt.show()



"""
Geomedian widget: generates an interactive visualisation of
the geomedian summary statistic.
"""

# Load modules
import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from odc.algo import xr_geomedian


def run_app():
    """
    An interactive app that allows users to visualise the difference between the median and geomedian
    time-series summary statistics. By modifying the red-green-blue values of three timesteps for a given pixel,
    the user changes the output summary statistics.

    This allows a visual representation of the difference through the output values,
    RGB colour, as well as showing values plotted as a vector on a 3-dimensional  space.

    Last modified: December 2021
    """

    # Define the red-green-blue sliders for timestep 1
    p1r = widgets.IntSlider(description="Red", max=255, value=58)
    p1g = widgets.IntSlider(description="Green", max=255, value=153)
    p1b = widgets.IntSlider(description="Blue", max=255, value=68)

    # Define the red-green-blue sliders for timestep 2
    p2r = widgets.IntSlider(description="Red", max=255, value=208)
    p2g = widgets.IntSlider(description="Green", max=255, value=221)
    p2b = widgets.IntSlider(description="Blue", max=255, value=203)

    # Define the red-green-blue sliders for timestep 3
    p3r = widgets.IntSlider(description="Red", max=255, value=202)
    p3g = widgets.IntSlider(description="Green", max=255, value=82)
    p3b = widgets.IntSlider(description="Blue", max=255, value=33)

    # Define the median calculation for the timesteps
    def f(p1r, p1g, p1b, p2r, p2g, p2b, p3r, p3g, p3b):
        print("Red Median = {}".format(np.median([p1r, p2r, p3r])))
        print("Green Median = {}".format(np.median([p1g, p2g, p3g])))
        print("Blue Median = {}".format(np.median([p1b, p2b, p3b])))

    # Define the geomedian calculation for the timesteps
    def g(p1r, p1g, p1b, p2r, p2g, p2b, p3r, p3g, p3b):
        print(
            "Red Geomedian = {:.2f}".format(
                xr_geomedian(
                    xr.Dataset(
                        {
                            "red": (
                                ("x", "y", "time"),
                                [[[np.float32(p1r), np.float32(p2r), np.float32(p3r)]]],
                            ),
                            "green": (
                                ("x", "y", "time"),
                                [[[np.float32(p1g), np.float32(p2g), np.float32(p3g)]]],
                            ),
                            "blue": (
                                ("x", "y", "time"),
                                [[[np.float32(p1b), np.float32(p2b), np.float32(p3b)]]],
                            ),
                        }
                    )
                ).red.values.ravel()[0]
            )
        )
        print(
            "Green Geomedian = {:.2f}".format(
                xr_geomedian(
                    xr.Dataset(
                        {
                            "red": (
                                ("x", "y", "time"),
                                [[[np.float32(p1r), np.float32(p2r), np.float32(p3r)]]],
                            ),
                            "green": (
                                ("x", "y", "time"),
                                [[[np.float32(p1g), np.float32(p2g), np.float32(p3g)]]],
                            ),
                            "blue": (
                                ("x", "y", "time"),
                                [[[np.float32(p1b), np.float32(p2b), np.float32(p3b)]]],
                            ),
                        }
                    )
                ).green.values.ravel()[0]
            )
        )
        print(
            "Blue Geomedian = {:.2f}".format(
                xr_geomedian(
                    xr.Dataset(
                        {
                            "red": (
                                ("x", "y", "time"),
                                [[[np.float32(p1r), np.float32(p2r), np.float32(p3r)]]],
                            ),
                            "green": (
                                ("x", "y", "time"),
                                [[[np.float32(p1g), np.float32(p2g), np.float32(p3g)]]],
                            ),
                            "blue": (
                                ("x", "y", "time"),
                                [[[np.float32(p1b), np.float32(p2b), np.float32(p3b)]]],
                            ),
                        }
                    )
                ).blue.values.ravel()[0]
            )
        )

    # Define the Timestep 1 box colour
    def h(p1r, p1g, p1b):
        fig1, axes1 = plt.subplots(figsize=(2, 2))
        fig1 = plt.imshow([[(p1r, p1g, p1b)]])
        axes1.set_title("Timestep 1")
        axes1.axis("off")
        plt.show(fig1)

    # Define the Timestep 2 box colour
    def hh(p2r, p2g, p2b):
        fig2, axes2 = plt.subplots(figsize=(2, 2))
        fig2 = plt.imshow([[(p2r, p2g, p2b)]])
        axes2.set_title("Timestep 2")
        axes2.axis("off")
        plt.show(fig2)

    # Define the Timestep 3 box colour
    def hhh(p3r, p3g, p3b):
        fig3, axes3 = plt.subplots(figsize=(2, 2))
        fig3 = plt.imshow([[(p3r, p3g, p3b)]])
        axes3.set_title("Timestep 3")
        axes3.axis("off")
        plt.show(fig3)

    # Define the Median RGB colour box
    def i(p1r, p1g, p1b, p2r, p2g, p2b, p3r, p3g, p3b):
        fig4, axes4 = plt.subplots(figsize=(3, 3))
        fig4 = plt.imshow(
            [
                [
                    (
                        int(np.median([p1r, p2r, p3r])),
                        int(np.median([p1g, p2g, p3g])),
                        int(np.median([p1b, p2b, p3b])),
                    )
                ]
            ]
        )
        axes4.set_title("Median RGB - All timesteps")
        axes4.axis("off")
        plt.show(fig4)

    # Define the Geomedian RGB colour box
    def ii(p1r, p1g, p1b, p2r, p2g, p2b, p3r, p3g, p3b):
        fig5, axes5 = plt.subplots(figsize=(3, 3))
        fig5 = plt.imshow(
            [
                [
                    (
                        int(
                            xr_geomedian(
                                xr.Dataset(
                                    {
                                        "red": (
                                            ("x", "y", "time"),
                                            [[[np.float32(p1r), np.float32(p2r), np.float32(p3r)]]],
                                        ),
                                        "green": (
                                            ("x", "y", "time"),
                                            [[[np.float32(p1g), np.float32(p2g), np.float32(p3g)]]],
                                        ),
                                        "blue": (
                                            ("x", "y", "time"),
                                            [[[np.float32(p1b), np.float32(p2b), np.float32(p3b)]]],
                                        ),
                                    }
                                )
                            ).red.values.ravel()[0]
                        ),
                        int(
                            xr_geomedian(
                                xr.Dataset(
                                    {
                                        "red": (
                                            ("x", "y", "time"),
                                            [[[np.float32(p1r), np.float32(p2r), np.float32(p3r)]]],
                                        ),
                                        "green": (
                                            ("x", "y", "time"),
                                            [[[np.float32(p1g), np.float32(p2g), np.float32(p3g)]]],
                                        ),
                                        "blue": (
                                            ("x", "y", "time"),
                                            [[[np.float32(p1b), np.float32(p2b), np.float32(p3b)]]],
                                        ),
                                    }
                                )
                            ).green.values.ravel()[0]
                        ),
                        int(
                            xr_geomedian(
                                xr.Dataset(
                                    {
                                        "red": (
                                            ("x", "y", "time"),
                                            [[[np.float32(p1r), np.float32(p2r), np.float32(p3r)]]],
                                        ),
                                        "green": (
                                            ("x", "y", "time"),
                                            [[[np.float32(p1g), np.float32(p2g), np.float32(p3g)]]],
                                        ),
                                        "blue": (
                                            ("x", "y", "time"),
                                            [[[np.float32(p1b), np.float32(p2b), np.float32(p3b)]]],
                                        ),
                                    }
                                )
                            ).blue.values.ravel()[0]
                        ),
                    )
                ]
            ]
        )
        axes5.set_title("Geomedian RGB - All timesteps")
        axes5.axis("off")
        plt.show(fig5)

    # Define 3-D axis to display vectors on
    def j(p1r, p1g, p1b, p2r, p2g, p2b, p3r, p3g, p3b):
        fig6 = plt.figure()
        axes6 = fig6.add_subplot(111, projection="3d")
        x = [
            p1r,
            p2r,
            p3r,
            int(np.median([p1r, p2r, p3r])),
            int(
                xr_geomedian(
                    xr.Dataset(
                        {
                            "red": (
                                ("x", "y", "time"),
                                [[[np.float32(p1r), np.float32(p2r), np.float32(p3r)]]],
                            ),
                            "green": (
                                ("x", "y", "time"),
                                [[[np.float32(p1g), np.float32(p2g), np.float32(p3g)]]],
                            ),
                            "blue": (
                                ("x", "y", "time"),
                                [[[np.float32(p1b), np.float32(p2b), np.float32(p3b)]]],
                            ),
                        }
                    )
                ).red.values.ravel()[0]
            ),
        ]
        y = [
            p1g,
            p2g,
            p3g,
            int(np.median([p1g, p2g, p3g])),
            int(
                xr_geomedian(
                    xr.Dataset(
                        {
                            "red": (
                                ("x", "y", "time"),
                                [[[np.float32(p1r), np.float32(p2r), np.float32(p3r)]]],
                            ),
                            "green": (
                                ("x", "y", "time"),
                                [[[np.float32(p1g), np.float32(p2g), np.float32(p3g)]]],
                            ),
                            "blue": (
                                ("x", "y", "time"),
                                [[[np.float32(p1b), np.float32(p2b), np.float32(p3b)]]],
                            ),
                        }
                    )
                ).green.values.ravel()[0]
            ),
        ]
        z = [
            p1b,
            p2b,
            p3b,
            int(np.median([p1b, p2b, p3b])),
            int(
                xr_geomedian(
                    xr.Dataset(
                        {
                            "red": (
                                ("x", "y", "time"),
                                [[[np.float32(p1r), np.float32(p2r), np.float32(p3r)]]],
                            ),
                            "green": (
                                ("x", "y", "time"),
                                [[[np.float32(p1g), np.float32(p2g), np.float32(p3g)]]],
                            ),
                            "blue": (
                                ("x", "y", "time"),
                                [[[np.float32(p1b), np.float32(p2b), np.float32(p3b)]]],
                            ),
                        }
                    )
                ).blue.values.ravel()[0]
            ),
        ]
        labels = [" 1", " 2", " 3", " median", " geomedian"]
        axes6.scatter(x, y, z, c=["black", "black", "black", "r", "blue"], marker="o")
        axes6.set_xlabel("Red")
        axes6.set_ylabel("Green")
        axes6.set_zlabel("Blue")
        axes6.set_xlim3d(0, 255)
        axes6.set_ylim3d(0, 255)
        axes6.set_zlim3d(0, 255)
        for ax, ay, az, label in zip(x, y, z, labels):
            axes6.text(ax, ay, az, label)
        plt.title("Each band represents a dimension.")
        plt.show()

    # Define outputs
    outf = widgets.interactive_output(
        f,
        {
            "p1r": p1r,
            "p2r": p2r,
            "p3r": p3r,
            "p1g": p1g,
            "p2g": p2g,
            "p3g": p3g,
            "p1b": p1b,
            "p2b": p2b,
            "p3b": p3b,
        },
    )
    outg = widgets.interactive_output(
        g,
        {
            "p1r": p1r,
            "p2r": p2r,
            "p3r": p3r,
            "p1g": p1g,
            "p2g": p2g,
            "p3g": p3g,
            "p1b": p1b,
            "p2b": p2b,
            "p3b": p3b,
        },
    )

    outh = widgets.interactive_output(h, {"p1r": p1r, "p1g": p1g, "p1b": p1b})
    outhh = widgets.interactive_output(hh, {"p2r": p2r, "p2g": p2g, "p2b": p2b})
    outhhh = widgets.interactive_output(hhh, {"p3r": p3r, "p3g": p3g, "p3b": p3b})

    outi = widgets.interactive_output(
        i,
        {
            "p1r": p1r,
            "p2r": p2r,
            "p3r": p3r,
            "p1g": p1g,
            "p2g": p2g,
            "p3g": p3g,
            "p1b": p1b,
            "p2b": p2b,
            "p3b": p3b,
        },
    )
    outii = widgets.interactive_output(
        ii,
        {
            "p1r": p1r,
            "p2r": p2r,
            "p3r": p3r,
            "p1g": p1g,
            "p2g": p2g,
            "p3g": p3g,
            "p1b": p1b,
            "p2b": p2b,
            "p3b": p3b,
        },
    )

    outj = widgets.interactive_output(
        j,
        {
            "p1r": p1r,
            "p2r": p2r,
            "p3r": p3r,
            "p1g": p1g,
            "p2g": p2g,
            "p3g": p3g,
            "p1b": p1b,
            "p2b": p2b,
            "p3b": p3b,
        },
    )

    app_output = widgets.HBox(
        [
            widgets.VBox(
                [
                    widgets.HBox([outh, widgets.VBox([p1r, p1g, p1b])]),
                    widgets.HBox([outhh, widgets.VBox([p2r, p2g, p2b])]),
                    widgets.HBox([outhhh, widgets.VBox([p3r, p3g, p3b])]),
                ]
            ),
            widgets.VBox(
                [widgets.HBox([widgets.VBox([outf, outi]), widgets.VBox([outg, outii])]), outj]
            ),
        ]
    )

    return app_output



"""
Coastal analyses on Digital Earth Africa data.
"""

import geopandas as gpd
import pandas as pd
import requests
from owslib.wfs import WebFeatureService
from pandas.plotting import register_matplotlib_converters
from shapely.geometry import box

# Fix converters for tidal plot
register_matplotlib_converters()


# URL for the DE Africa Coastlines data on Geoserver.
WFS_ADDRESS = "https://geoserver.digitalearth.africa/geoserver/wfs"


def model_tides(*args, **kwargs):
    raise ImportError(
        "The `model_tides` function has been removed and is no longer available in this package.\n"
        "Please install and use the `eo-tides` package instead:\n"
        "https://geoscienceaustralia.github.io/eo-tides/migration/"
    )


def pixel_tides(*args, **kwargs):
    raise ImportError(
        "The `pixel_tides` function has been removed and is no longer available in this package.\n"
        "Please install and use the `eo-tides` package instead:\n"
        "https://geoscienceaustralia.github.io/eo-tides/migration/"
    )


def tidal_tag(*args, **kwargs):
    raise ImportError(
        "The `tidal_tag` function has been removed and is no longer available in this package.\n"
        "Please install and use the `eo-tides` package instead:\n"
        "https://geoscienceaustralia.github.io/eo-tides/migration/"
    )


def tidal_stats(*args, **kwargs):
    raise ImportError(
        "The `tidal_stats` function has been removed and is no longer available in this package.\n"
        "Please install and use the `eo-tides` package instead:\n"
        "https://geoscienceaustralia.github.io/eo-tides/migration/"
    )


def transect_distances(transects_gdf, lines_gdf, mode="distance"):
    """
    Take a set of transects (e.g. shore-normal beach survey lines), and
    determine the distance along the transect to each object in a set of
    lines (e.g. shorelines). Distances are measured in the CRS of the
    input datasets.

    For coastal applications, transects should be drawn from land to
    water (with the first point being on land so that it can be used
    as a consistent location from which to measure distances.

    The distance calculation can be performed using two modes:
        - 'distance': Distances are measured from the start of the
          transect to where it intersects with each line. Any transect
          that intersects a line more than once is ignored. This mode is
          useful for measuring e.g. the distance to the shoreline over
          time from a consistent starting location.
        - 'width' Distances are measured between the first and last
          intersection between a transect and each line. Any transect
          that intersects a line only once is ignored. This is useful
          for e.g. measuring the width of a narrow area of coastline over
          time, e.g. the neck of a spit or tombolo.

    Parameters
    ----------
    transects_gdf : geopandas.GeoDataFrame
        A GeoDataFrame containing one or multiple vector profile lines.
        The GeoDataFrame's index column will be used to name the rows in
        the output distance table.
    lines_gdf : geopandas.GeoDataFrame
        A GeoDataFrame containing one or multiple vector line features
        that intersect the profile lines supplied to `transects_gdf`.
        The GeoDataFrame's index column will be used to name the columns
        in the output distance table.
    mode : string, optional
        Whether to use 'distance' (for measuring distances from the
        start of a profile) or 'width' mode (for measuring the width
        between two profile intersections). See docstring above for more
        info; defaults to 'distance'.

    Returns
    -------
    distance_df : pandas.DataFrame
        A DataFrame containing distance measurements for each profile
        line (rows) and line feature (columns).
    """

    import warnings

    from shapely.errors import ShapelyDeprecationWarning
    from shapely.geometry import Point

    def _intersect_dist(transect_gdf, lines_gdf, mode=mode):
        """
        Take an individual transect, and determine the distance along
        the transect to each object in a set of lines (e.g. shorelines).
        """

        # Identify intersections between transects and lines
        intersect_points = lines_gdf.apply(
            lambda x: x.geometry.intersection(transect_gdf.geometry), axis=1
        )

        # In distance mode, identify transects with one intersection only,
        # and use this as the end point and the start of the transect as the
        # start point when measuring distances
        if mode == "distance":
            start_point = Point(transect_gdf.geometry.coords[0])
            point_df = intersect_points.apply(
                lambda x: (
                    pd.Series({"start": start_point, "end": x})
                    if x.type == "Point"
                    else pd.Series({"start": None, "end": None})
                )
            )

        # In width mode, identify transects with multiple intersections, and
        # use the first intersection as the start point and the second
        # intersection for the end point when measuring distances
        if mode == "width":
            point_df = intersect_points.apply(
                lambda x: (
                    pd.Series({"start": x.geoms[0], "end": x.geoms[-1]})
                    if x.type == "MultiPoint"
                    else pd.Series({"start": None, "end": None})
                )
            )

        # Calculate distances between valid start and end points
        distance_df = point_df.apply(lambda x: x.start.distance(x.end) if x.start else None, axis=1)

        return distance_df

    # Run code after ignoring Shapely pre-v2.0 warnings
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)

        # Assert that both datasets use the same CRS
        assert (
            transects_gdf.crs == lines_gdf.crs
        ), "Please ensure both input datasets use the same CRS."

        # Run distance calculations
        distance_df = transects_gdf.apply(lambda x: _intersect_dist(x, lines_gdf), axis=1)

        return pd.DataFrame(distance_df)


def get_coastlines(
    bbox: tuple, crs="EPSG:4326", layer="shorelines", drop_wms=True
) -> gpd.GeoDataFrame:
    """
    Get DE Africa Coastlines data for a provided bounding box using WFS.

    For a full description of the DE Africa Coastlines dataset, refer to the
    official Digital Earth Africa product description:

    Parameters
    ----------
    bbox : (xmin, ymin, xmax, ymax), or geopandas object
        Bounding box expressed as a tuple. Alternatively, a bounding
        box can be automatically extracted by suppling a
        geopandas.GeoDataFrame or geopandas.GeoSeries.
    crs : str, optional
        Optional CRS for the bounding box. This is ignored if `bbox`
        is provided as a geopandas object.
    layer : str, optional
        Which DE Africa Coastlines layer to load. Options include the annual
        shoreline vectors ("shorelines") and the rates of change
        statistics points ("statistics"). Defaults to "shorelines".
    drop_wms : bool, optional
        Whether to drop WMS-specific attribute columns from the data.
        These columns are used for visualising the dataset on DE Africa Maps,
        and are unlikely to be useful for scientific analysis. Defaults
        to True.

    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame containing shoreline or point features and
        associated metadata.
    """

    # If bbox is a geopandas object, convert to bbox.
    try:
        crs = str(bbox.crs)
        bbox = bbox.total_bounds
    except Exception:
        pass

    # Get the available layers in the coastlines:DEAfrica_Coastlines group.
    describe_layer_url = (
        "https://geoserver.digitalearth.africa/geoserver/wms?service=WMS&version=1.1.1&"
        "request=DescribeLayer&layers=coastlines:DEAfrica_Coastlines&outputFormat=application/json"
    )
    describe_layer_response = requests.get(describe_layer_url).json()
    available_layers = [
        layer["layerName"] for layer in describe_layer_response["layerDescriptions"]
    ]

    # Get the layer name.
    if layer == "shorelines":
        layer_name = [i for i in available_layers if "shorelines" in i]
    else:
        layer_name = [i for i in available_layers if "rates_of_change" in i]

    # Query WFS.
    wfs = WebFeatureService(url=WFS_ADDRESS, version="1.1.0")
    response = wfs.getfeature(typename=layer_name, bbox=tuple(bbox) + (crs,), outputFormat="json")

    # Load data as a geopandas.GeoDataFrame.
    coastlines_gdf = gpd.read_file(response)

    # Clip to extent of bounding box.
    extent = gpd.GeoSeries(box(*bbox), crs=crs).to_crs(coastlines_gdf.crs)
    coastlines_gdf = coastlines_gdf.clip(extent)

    # Optionally drop WMS-specific columns.
    if drop_wms:
        coastlines_gdf = coastlines_gdf.loc[:, ~coastlines_gdf.columns.str.contains("wms_")]

    return coastlines_gdf



# crophealth.py
"""
Functions for loading and interacting with data in the crop health notebook,
 inside the Real_world_examples folder.
"""

# Load modules

# Force GeoPandas to use Shapely instead of PyGEOS
# In a future release, GeoPandas will switch to using Shapely by default.

import os

os.environ["USE_PYGEOS"] = "0"

import datetime as dt
import json
import warnings
from io import BytesIO

import datacube
import geopandas as gpd
import ipywidgets as widgets
import matplotlib as mpl
import matplotlib.pyplot as plt
import xarray as xr
from ipyleaflet import DrawControl, GeoJSON, Map, basemaps
from IPython.display import display
from osgeo import ogr

from deafrica_tools.bandindices import calculate_indices
from deafrica_tools.datahandling import load_ard
from deafrica_tools.spatial import xr_rasterize


def load_crophealth_data(lat: float, lon: float, buffer: float, date: str) -> xr.Dataset:
    """
    Loads Sentinel-2 analysis-ready data (ARD) product for the crop health
    case-study area over the last two years.
    Last modified: April 2020

    Parameters
    ----------
    lat: float
        The central latitude to analyse
    lon: float
        The central longitude to analyse
    buffer:
         The number of square degrees to load around the central latitude and longitude.
         For reasonable loading times, set this as `0.1` or lower.
    date:
         The most recent date to show data for.
         The app will automatically load all data available for the two years prior to this date.

    Returns
    ----------
    ds: xarray.Dataset
        data set containing combined, masked data
        Masked values are set to 'nan'
    """

    # Suppress warnings
    warnings.filterwarnings("ignore")

    # Initialise the data cube. 'app' argument is used to identify this app
    dc = datacube.Datacube(app="Crophealth-app")

    # Define area to load
    latitude = (lat - buffer, lat + buffer)
    longitude = (lon - buffer, lon + buffer)

    # Specify the date range
    # Calculated as today's date, subtract 730 days to collect two years of data
    # Dates are converted to strings as required by loading function below
    end_date = dt.datetime.strptime(date, "%Y-%m-%d")
    start_date = end_date - dt.timedelta(days=730)

    time = (start_date.strftime("%Y-%m-%d"), end_date.strftime("%Y-%m-%d"))

    # Construct the data cube query
    products = ["s2_l2a"]

    query = {
        "x": longitude,
        "y": latitude,
        "time": time,
        "measurements": ["red", "green", "blue", "nir", "swir_2"],
        "output_crs": "EPSG:6933",
        "resolution": (-20, 20),
    }

    # Load the data and mask out bad quality pixels
    ds = load_ard(dc, products=products, min_gooddata=0.5, **query)

    # Calculate the normalised difference vegetation index (NDVI) across
    # all pixels for each image.
    # This is stored as an attribute of the data
    ds = calculate_indices(ds, index="NDVI", satellite_mission="s2")

    # Return the data
    return ds


def run_crophealth_app(ds: xr.Dataset, lat: float, lon: float, buffer: float):
    """
    Plots an interactive map of the crop health case-study area and allows
    the user to draw polygons. This returns a plot of the average NDVI value
    in the polygon area.
    Last modified: January 2020

    Parameters
    ----------
    ds: xarray.Dataset
        data set containing combined, masked data
        Masked values are set to 'nan'
    lat: float
        The central latitude corresponding to the area of loaded ds
    lon: float
        The central longitude corresponding to the area of loaded ds
    buffer:
         The number of square degrees to load around the central latitude and longitude.
         For reasonable loading times, set this as `0.1` or lower.
    """

    # Suppress warnings
    warnings.filterwarnings("ignore")

    # Update plotting functionality through rcParams
    mpl.rcParams.update({"figure.autolayout": True})

    # Define polygon bounds
    latitude = (lat - buffer, lat + buffer)
    longitude = (lon - buffer, lon + buffer)

    # Define the bounding box that will be overlayed on the interactive map
    # The bounds are hard-coded to match those from the loaded data
    geom_obj = {
        "type": "Feature",
        "properties": {
            "style": {
                "stroke": True,
                "color": "red",
                "weight": 4,
                "opacity": 0.8,
                "fill": True,
                "fillColor": False,
                "fillOpacity": 0,
                "showArea": True,
                "clickable": True,
            }
        },
        "geometry": {
            "type": "Polygon",
            "coordinates": [
                [
                    [longitude[0], latitude[0]],
                    [longitude[1], latitude[0]],
                    [longitude[1], latitude[1]],
                    [longitude[0], latitude[1]],
                    [longitude[0], latitude[0]],
                ]
            ],
        },
    }

    # Create a map geometry from the geom_obj dictionary
    # center specifies where the background map view should focus on
    # zoom specifies how zoomed in the background map should be
    loadeddata_geometry = ogr.CreateGeometryFromJson(str(geom_obj["geometry"]))
    loadeddata_center = [
        loadeddata_geometry.Centroid().GetY(),
        loadeddata_geometry.Centroid().GetX(),
    ]
    loadeddata_zoom = 16

    # define the study area map
    studyarea_map = Map(
        center=loadeddata_center, zoom=loadeddata_zoom, basemap=basemaps.Esri.WorldImagery
    )

    # define the drawing controls
    studyarea_drawctrl = DrawControl(
        polygon={"shapeOptions": {"fillOpacity": 0}},
        marker={},
        circle={},
        circlemarker={},
        polyline={},
    )

    # add drawing controls and data bound geometry to the map
    studyarea_map.add(studyarea_drawctrl)
    studyarea_map.add(GeoJSON(data=geom_obj))

    # Index to count drawn polygons
    polygon_number = 0

    # Define widgets to interact with
    instruction = widgets.Output(layout={"border": "1px solid black"})
    with instruction:
        print(
            "Draw a polygon within the red box to view a plot of "
            "average NDVI over time in that area."
        )

    info = widgets.Output(layout={"border": "1px solid black"})
    with info:
        print("Plot status:")

    fig_display = widgets.Output(
        layout=widgets.Layout(
            width="50%",  # proportion of horizontal space taken by plot
        )
    )

    with fig_display:
        plt.ioff()
        fig, ax = plt.subplots(figsize=(8, 6))
        ax.set_ylim([0, 1])

    colour_list = plt.rcParams["axes.prop_cycle"].by_key()["color"]

    # Function to execute each time something is drawn on the map
    def handle_draw(self, action, geo_json):
        nonlocal polygon_number

        # Execute behaviour based on what the user draws
        if geo_json["geometry"]["type"] == "Polygon":

            info.clear_output(wait=True)  # wait=True reduces flicker effect

            # Save geojson polygon to io temporary file to be rasterized later
            jsonData = json.dumps(geo_json)
            binaryData = jsonData.encode()
            io = BytesIO(binaryData)
            io.seek(0)

            # Read the polygon as a geopandas dataframe
            gdf = gpd.read_file(io)
            gdf.crs = "EPSG:4326"

            # Convert the drawn geometry to pixel coordinates
            xr_poly = xr_rasterize(gdf, ds.NDVI.isel(time=0), crs="EPSG:6933")

            # Construct a mask to only select pixels within the drawn polygon
            masked_ds = ds.NDVI.where(xr_poly)

            masked_ds_mean = masked_ds.mean(dim=["x", "y"], skipna=True)
            colour = colour_list[polygon_number % len(colour_list)]

            # Add a layer to the map to make the most recently drawn polygon
            # the same colour as the line on the plot
            studyarea_map.add(
                GeoJSON(
                    data=geo_json,
                    style={"color": colour, "opacity": 1, "weight": 4.5, "fillOpacity": 0.0},
                )
            )

            # add new data to the plot
            xr.plot.plot(masked_ds_mean, marker="*", color=colour, ax=ax)

            # reset titles back to custom
            ax.set_title("Average NDVI from Sentinel-2")
            ax.set_xlabel("Date")
            ax.set_ylabel("NDVI")

            # refresh display
            fig_display.clear_output(wait=True)  # wait=True reduces flicker effect
            with fig_display:
                display(fig)

            with info:
                print("Plot status: polygon sucessfully added to plot.")

            # Iterate the polygon number before drawing another polygon
            polygon_number = polygon_number + 1

        else:
            info.clear_output(wait=True)
            with info:
                print(
                    "Plot status: this drawing tool is not currently "
                    "supported. Please use the polygon tool."
                )

    # call to say activate handle_draw function on draw
    studyarea_drawctrl.on_draw(handle_draw)

    with fig_display:
        # TODO: update with user friendly something
        display(widgets.HTML(""))

    # Construct UI:
    #  +-----------------------+
    #  | instruction           |
    #  +-----------+-----------+
    #  |  map      |  plot     |
    #  |           |           |
    #  +-----------+-----------+
    #  | info                  |
    #  +-----------------------+
    ui = widgets.VBox([instruction, widgets.HBox([studyarea_map, fig_display]), info])
    display(ui)



# -*- coding: utf-8 -*-
"""
Satellite imagery animation widget, which can be used to interactively
produce animations for multiple DE Africa products.
"""
import datetime
import itertools
import json
import warnings
from io import BytesIO

import datacube
import geopandas as gpd
import ipywidgets as widgets
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from datacube.utils.geometry import Geometry
from datacube.utils.masking import mask_invalid_data
from ipyleaflet import LayerGroup, basemap_to_tiles, basemaps
from ipywidgets import HTML, Button, GridspecLayout, HBox, Layout, Output, VBox
from skimage.exposure import rescale_intensity
from skimage.filters import unsharp_mask

import deafrica_tools.app.widgetconstructors as deawidgets
from deafrica_tools.coastal import get_coastlines
from deafrica_tools.dask import create_local_dask_cluster
from deafrica_tools.datahandling import load_ard
from deafrica_tools.plotting import xr_animation
from deafrica_tools.spatial import reverse_geocode

warnings.filterwarnings("ignore")


# WMS params and satellite style bands
sat_params = {
    "Landsat": {
        "products": ["ls5_sr", "ls7_sr", "ls8_sr", "ls9_sr"],
        "styles": {
            "True colour": ("true_colour", ["red", "green", "blue"]),
            "False colour": (
                "false_colour",
                ["swir_1", "nir", "green"],
            ),
        },
    },
    "Sentinel-2": {
        "products": ["s2_l2a"],
        "styles": {
            "True colour": ("simple_rgb", ["red", "green", "blue"]),
            "False colour": (
                "infrared_green",
                ["swir_2", "nir_1", "green"],
            ),
        },
    },
}


def make_box_layout():
    return Layout(
        #         border='solid 1px black',
        margin="0px 10px 10px 0px",
        padding="5px 5px 5px 5px",
        width="100%",
        height="100%",
    )


def create_expanded_button(description, button_style):
    return Button(
        description=description,
        button_style=button_style,
        layout=Layout(width="auto", height="auto"),
    )


def update_map_layers(self):
    """
    Updates map to add new DE Africa layers, styles or basemap when selected
    using menu options. Triggers data reload by resetting load params
    and output arrays.
    """

    # Clear data load params to trigger data re-load
    self.timeseries_ds = None
    self.load_params = None
    self.query_params = None

    # Clear all layers and add basemap
    self.map_layers.clear_layers()
    self.map_layers.add_layer(self.basemap)


def extract_data(self):

    # Connect to datacube database
    dc = datacube.Datacube(app="Exporting satellite images")

    # Configure local dask cluster
    client = create_local_dask_cluster(return_client=True, display_client=True)

    # Convert to geopolygon
    geopolygon = Geometry(geom=self.gdf_drawn.geometry[0], crs=self.gdf_drawn.crs)

    # Create query.
    start_date = np.datetime64(self.start_date)
    end_date = np.datetime64(self.end_date)

    self.query_params = {
        "time": (str(start_date), str(end_date)),
        "geopolygon": geopolygon,
    }

    # Find matching datasets
    dss = [
        dc.find_datasets(product=i, **self.query_params)
        for i in sat_params[self.dealayer]["products"]
    ]
    dss = list(itertools.chain.from_iterable(dss))

    # If data is found
    if len(dss) > 0:

        # Get CRS
        crs = str(dss[0].crs)

        self.load_params = {
            "measurements": sat_params[self.dealayer]["styles"][self.style][1],
            "resolution": (-self.resolution, self.resolution),
            "output_crs": crs,
            "group_by": "solar_day",
            "dask_chunks": {"time": 1, "x": 2048, "y": 2048},
            "resampling": {"*": "cubic", "oa_fmask": "nearest", "fmask": "nearest"},
        }

        # Load data

        timeseries_ds = load_ard(
            dc=dc,
            products=sat_params[self.dealayer]["products"],
            min_gooddata=1.0 - (self.max_cloud_cover / 100),
            ls7_slc_off=False,
            mask_pixel_quality=self.cloud_mask,
            **self.load_params,
            **self.query_params,
        )

        # Set invalid nodata pixels to NaN
        timeseries_ds = mask_invalid_data(timeseries_ds)

    # Else if no data is returned, return None
    else:
        timeseries_ds = None

    # Close down the dask client
    client.close()

    return timeseries_ds.compute()


def plot_data(self, fname):

    # Data to plot
    to_plot = self.timeseries_ds

    # If rolling median specified
    if self.rolling_median:
        with self.status_info:
            print(f"\nApplying rolling median ({self.rolling_median_window} timesteps window)")
        to_plot = to_plot.rolling(
            time=int(self.rolling_median_window), center=True, min_periods=1
        ).median()

    # If resampling freq specified
    if self.resample_freq:
        with self.status_info:
            print(f"\nResampling data to {self.resample_freq} frequency")
        to_plot = to_plot.resample(time=self.resample_freq).median()

    # Raise by power to dampen bright features and enhance dark.
    # Raise vmin and vmax by same amount to ensure proper stretch
    if self.power < 1.0:
        with self.status_info:
            print(f"\nApplying power transformation ({self.power})")
        to_plot = to_plot**self.power

    # Apply unsharp masking to enhance overall dynamic range,
    # and improve fine scale detail
    if self.unsharp_mask:
        with self.status_info:
            print(
                f"\nApplying unsharp masking with {self.unsharp_mask_radius} "
                f"radius and {self.unsharp_mask_amount} amount"
            )

        funcs_list = [
            rescale_intensity,
            lambda x: unsharp_mask(
                x, radius=self.unsharp_mask_radius, amount=self.unsharp_mask_amount
            ),
        ]
    else:
        funcs_list = None

    xr_animation(
        output_path=fname,
        ds=to_plot.dropna(dim="time", how="all"),
        show_text="",
        bands=sat_params[self.dealayer]["styles"][self.style][1],
        interval=self.interval,
        width_pixels=self.width,
        show_gdf=deacoastlines_overlay(to_plot) if self.deacoastlines else None,
        gdf_kwargs={"linewidth": 3},
        percentile_stretch=(self.vmin, self.vmax),
        image_proc_funcs=funcs_list,
        show_date="%Y" if self.resample_freq == "1Y" else "%b %Y",
        annotation_kwargs={"fontsize": 75},
    )

    # Add plot preview below map and finish
    plt.show()
    with self.status_info:
        print(f"\nImage successfully exported to:\n{fname}.")


def deacoastlines_overlay(ds):
    # Get bounding box of data
    xmin, ymin, xmax, ymax = ds.geobox.geographic_extent.boundingbox
    bounds = (xmin, ymin, xmax, ymax)

    # Load data
    deacl_gdf = get_coastlines(bbox=bounds)

    # Clip to extent of satellite data
    bbox = gpd.GeoDataFrame(geometry=[ds.geobox.extent.geom], crs=ds.geobox.crs)
    deacl_gdf = gpd.overlay(deacl_gdf, bbox.to_crs(deacl_gdf.crs))

    if len(deacl_gdf.index) > 0:
        deacl_gdf = deacl_gdf.dissolve("year")  # values("year", ascending=True)
        # Apply colours
        norm = matplotlib.colors.Normalize(vmin=0, vmax=len(deacl_gdf.index))
        cmap = matplotlib.cm.get_cmap("inferno")
        rgba = cmap(norm(deacl_gdf.reset_index().index))
        deacl_gdf["color"] = list(rgba)
        deacl_gdf["start_time"] = pd.to_datetime(deacl_gdf.index) + pd.DateOffset(months=0)
        deacl_gdf = deacl_gdf.sort_index()
        return deacl_gdf
    else:
        print(f"No annual shorelines data available for bounding box {bounds}")
        return None


class animation_app(HBox):
    def __init__(self):
        super().__init__()

        ######################
        # INITIAL ATTRIBUTES #
        ######################

        # Basemap
        self.basemap_list = [
            ("ESRI World Imagery", basemap_to_tiles(basemaps.Esri.WorldImagery)),
            ("Open Street Map", basemap_to_tiles(basemaps.OpenStreetMap.Mapnik)),
        ]
        self.basemap = self.basemap_list[0][1]

        # Satellite data
        end_date = datetime.datetime.today()
        start_date = datetime.datetime(
            year=end_date.year - 3, month=end_date.month, day=end_date.day
        )
        self.start_date = start_date.strftime("%Y-%m-%d")
        self.end_date = end_date.strftime("%Y-%m-%d")
        self.dealayer_list = [
            ("Landsat", "Landsat"),
            ("Sentinel-2", "Sentinel-2"),
        ]
        self.dealayer = self.dealayer_list[0][1]

        # Styles
        self.styles_list = ["True colour", "False colour"]
        self.style = self.styles_list[0]

        # Analysis params
        self.resolution = 30
        self.vmin = 0.01
        self.vmax = 0.99
        self.power = 1.0
        self.output_list = [("MP4", "mp4"), ("GIF", "gif")]
        self.output_format = self.output_list[0][1]
        self.rolling_median = False
        self.rolling_median_window = 20
        self.unsharp_mask = False
        self.unsharp_mask_radius = 20
        self.unsharp_mask_amount = 0.3
        self.max_size = False
        self.width = 900
        self.interval = 100
        self.cloud_mask = False
        self.max_cloud_cover = 20
        self.resample_list = [
            ("None", False),
            ("Monthly", "1M"),
            ("Quarterly", "Q-DEC"),
            ("Yearly", "1Y"),
        ]
        self.resample_freq = self.resample_list[0][1]
        self.deacoastlines = False

        # Drawing params
        self.target = None
        self.action = None
        self.gdf_drawn = None

        # Data load params
        self.timeseries_ds = None
        self.load_params = None
        self.query_params = None

        ##################
        # HEADER FOR APP #
        ##################

        # Create the Header widget
        header_title_text = "<h3>Digital Earth Africa satellite imagery animations</h3>"
        instruction_text = (
            "<p>Select the desired satellite data, imagery date range "
            "and image style, then zoom in and draw a rectangle to "
            "select an area export as a satellite imagery time-series "
            "animation.</p>"
        )
        self.header = deawidgets.create_html(f"{header_title_text}{instruction_text}")
        self.header.layout = make_box_layout()

        #####################################
        # HANDLER FUNCTION FOR DRAW CONTROL #
        #####################################

        # Define the action to take once something is drawn on the map
        def update_geojson(target, action, geo_json):

            # Get data from action
            self.action = action

            # Clear data load params to trigger data re-load
            self.timeseries_ds = None
            self.load_params = None
            self.query_params = None

            # Convert data to geopandas
            json_data = json.dumps(geo_json)
            binary_data = json_data.encode()
            io = BytesIO(binary_data)
            io.seek(0)
            gdf = gpd.read_file(io)
            gdf.crs = "EPSG:4326"

            # Convert to WGS 84 / NSIDC EASE-Grid 2.0 Global and compute area
            gdf_drawn_nsidc = gdf.copy().to_crs("EPSG:6933")
            m2_per_ha = 10000
            area = gdf_drawn_nsidc.area.values[0] / m2_per_ha
            polyarea_label = "Total area of satellite data to extract"
            polyarea_text = f"<b>{polyarea_label}</b>: {area:.2f} ha</sup>"

            # Test area size
            if self.max_size:
                confirmation_text = (
                    '<span style="color: #33cc33"> '
                    "<b>(Overriding maximum size limit; use with caution as may lead to memory issues)</b></span>"
                )
                self.header.value = (
                    header_title_text + instruction_text + polyarea_text + confirmation_text
                )
                self.gdf_drawn = gdf
            elif area <= 50000:
                confirmation_text = (
                    '<span style="color: #33cc33"> '
                    "<b>(Area to extract falls within "
                    "recommended 50000 ha limit)</b></span>"
                )
                self.header.value = (
                    header_title_text + instruction_text + polyarea_text + confirmation_text
                )
                self.gdf_drawn = gdf
            else:
                warning_text = (
                    '<span style="color: #ff5050"> '
                    "<b>(Area to extract is too large, "
                    "please select an area less than 50000 "
                    "ha)</b></span>"
                )
                self.header.value = (
                    header_title_text + instruction_text + polyarea_text + warning_text
                )
                self.gdf_drawn = None

        ###########################
        # WIDGETS FOR APP OUTPUTS #
        ###########################

        self.status_info = Output(layout=make_box_layout())
        self.output_plot = Output(layout=make_box_layout())

        #########################################
        # MAP WIDGET, DRAWING TOOLS, WMS LAYERS #
        #########################################

        # Create drawing tools
        desired_drawtools = ["rectangle"]
        draw_control = deawidgets.create_drawcontrol(desired_drawtools)

        # Begin by displaying an empty layer group, and update the group with desired WMS on interaction.
        self.map_layers = LayerGroup(layers=())
        self.map_layers.name = "Map Overlays"

        # Create map widget
        self.m = deawidgets.create_map(map_center=(5.65, 26.17), zoom_level=13)
        self.m.layout = make_box_layout()

        # Add tools to map widget
        self.m.add_control(draw_control)
        self.m.add_layer(self.map_layers)

        # Update all maps to starting defaults
        update_map_layers(self)

        ############################
        # WIDGETS FOR APP CONTROLS #
        ############################

        # Create parameter widgets
        dropdown_basemap = deawidgets.create_dropdown(self.basemap_list, self.basemap_list[0][1])
        dropdown_dealayer = deawidgets.create_dropdown(self.dealayer_list, self.dealayer_list[0][1])
        dropdown_output = deawidgets.create_dropdown(self.output_list, self.output_list[0][1])
        date_picker_start = deawidgets.create_datepicker(
            value=start_date,
        )
        date_picker_end = deawidgets.create_datepicker(
            value=end_date,
        )
        dropdown_styles = deawidgets.create_dropdown(self.styles_list, self.styles_list[0])
        slider_percentile = widgets.FloatRangeSlider(
            value=[0.01, 0.99],
            min=0,
            max=1,
            step=0.001,
            description="",
            layout={"width": "85%"},
        )
        run_button = create_expanded_button("Generate animation", "info")

        floatslider_max_cloud_cover = widgets.IntSlider(
            value=20,
            min=0,
            max=100,
            step=1,
            description="",
            layout={"width": "85%"},
        )

        checkbox_rolling_median = deawidgets.create_checkbox(
            self.rolling_median,
            "Apply rolling median to produce smooth, cloud-free animations",
            layout={"width": "90%", "height": "4em"},
        )
        text_rolling_median_window = widgets.IntText(
            value=20,
            step=1,
            description="Rolling window (timesteps)",
            layout={
                "width": "85%",
                "margin": "0px",
                "padding": "0px",
                "display": "none",
            },
        )

        # Expandable advanced section
        text_interval = widgets.IntText(value=100, description="", step=50, layout={"width": "95%"})
        text_resolution = widgets.FloatText(
            value=30,
            description="",
            layout={"width": "95%", "margin": "0px", "padding": "0px"},
        )
        text_width = widgets.IntText(value=900, description="", step=50, layout={"width": "95%"})
        dropdown_resampling = deawidgets.create_dropdown(
            self.resample_list,
            self.resample_freq,
            description="",
            layout={"width": "95%"},
        )
        checkbox_cloud_mask = deawidgets.create_checkbox(
            self.cloud_mask,
            "Mask out cloudy pixels",
            layout={"width": "95%", "height": "auto"},
        )
        slider_power = widgets.FloatSlider(
            value=1.0,
            min=0.01,
            max=1.0,
            step=0.01,
            description="",
            layout={"width": "95%"},
        )
        checkbox_unsharp_mask = deawidgets.create_checkbox(
            self.unsharp_mask, "Enable", layout={"width": "95%"}
        )
        text_unsharp_mask_radius = widgets.FloatText(
            value=20,
            step=1,
            description="Radius",
            layout={
                "width": "95%",
                "margin": "0px",
                "padding": "0px",
                "display": "none",
            },
        )
        text_unsharp_mask_amount = widgets.FloatText(
            value=0.3,
            step=0.1,
            description="Amount",
            layout={
                "width": "95%",
                "margin": "0px",
                "padding": "0px",
                "display": "none",
            },
        )
        checkbox_deacoastlines = deawidgets.create_checkbox(
            self.deacoastlines, "Add DE Africa Coastlines overlay", layout={"width": "95%"}
        )
        checkbox_max_size = deawidgets.create_checkbox(
            self.max_size, "Enable", layout={"width": "95%"}
        )
        expand_box = widgets.VBox(
            [
                HTML("Frame interval (milliseconds):"),
                text_interval,
                HTML("</br>Resolution (metres):"),
                text_resolution,
                HTML("</br>Width of output animation in pixels:"),
                text_width,
                HTML("</br>Apply temporal resampling:"),
                dropdown_resampling,
                HTML("</br>"),
                checkbox_cloud_mask,
                checkbox_deacoastlines,
                HTML("</br>Apply power transformation to darken bright features:"),
                slider_power,
                HTML("</br>Apply unsharp masking to sharpen imagery:"),
                checkbox_unsharp_mask,
                text_unsharp_mask_radius,
                text_unsharp_mask_amount,
                HTML(
                    "</br>Override maximum size limit: (use with caution; may cause memory issues/crashes)"
                ),
                checkbox_max_size,
            ],
        )

        expand = widgets.Accordion(
            children=[expand_box],
            selected_index=None,
        )
        expand.set_title(0, "Advanced")

        # Add specific dialogs to class so they can be modified
        self.text_resolution = text_resolution
        self.text_unsharp_mask_radius = text_unsharp_mask_radius
        self.text_unsharp_mask_amount = text_unsharp_mask_amount
        self.text_rolling_median_window = text_rolling_median_window

        ####################################
        # UPDATE FUNCTIONS FOR EACH WIDGET #
        ####################################

        # Run update functions whenever various widgets are changed.
        date_picker_start.observe(self.update_start_date, "value")
        date_picker_end.observe(self.update_end_date, "value")
        dropdown_basemap.observe(self.update_basemap, "value")
        dropdown_dealayer.observe(self.update_dealayer, "value")
        dropdown_styles.observe(self.update_styles, "value")

        slider_percentile.observe(self.update_slider_percentile, "value")
        floatslider_max_cloud_cover.observe(self.update_floatslider_max_cloud_cover, "value")
        checkbox_rolling_median.observe(self.update_checkbox_rolling_median, "value")
        text_rolling_median_window.observe(self.update_text_rolling_median_window, "value")
        dropdown_output.observe(self.update_output, "value")
        run_button.on_click(self.run_app)
        draw_control.on_draw(update_geojson)

        # Advanced params
        text_resolution.observe(self.update_text_resolution, "value")
        slider_power.observe(self.update_slider_power, "value")
        text_width.observe(self.update_width, "value")
        text_interval.observe(self.update_interval, "value")
        dropdown_resampling.observe(self.update_dropdown_resampling, "value")
        checkbox_cloud_mask.observe(self.update_checkbox_cloud_mask, "value")
        checkbox_unsharp_mask.observe(self.update_checkbox_unsharp_mask, "value")
        text_unsharp_mask_radius.observe(self.update_text_unsharp_mask_radius, "value")
        text_unsharp_mask_amount.observe(self.update_text_unsharp_mask_amount, "value")
        checkbox_deacoastlines.observe(self.update_deacoastlines, "value")
        checkbox_max_size.observe(self.update_checkbox_max_size, "value")

        ##################################
        # COLLECTION OF ALL APP CONTROLS #
        ##################################

        parameter_selection = VBox(
            [
                HTML("<b>Satellite imagery:</b>"),
                dropdown_dealayer,
                HTML("<b>Start date:</b>"),
                date_picker_start,
                HTML("<b>End date:</b>"),
                date_picker_end,
                HTML("<b>Style:</b>"),
                dropdown_styles,
                HTML("<b>Colour percentile stretch:</b>"),
                slider_percentile,
                HTML("<b>Maximum cloud cover (%):</b>"),
                floatslider_max_cloud_cover,
                checkbox_rolling_median,
                text_rolling_median_window,
                HTML("</br><b>Output file format:</b>"),
                dropdown_output,
                HTML("</br>"),
                expand,
            ]
        )
        map_selection = VBox(
            [
                HTML("</br><b>Map overlay:</b>"),
                dropdown_basemap,
            ]
        )
        parameter_selection.layout = make_box_layout()
        map_selection.layout = make_box_layout()

        ###############################
        # SPECIFICATION OF APP LAYOUT #
        ###############################

        #       0   1    2   3   4   5   6   7    8   9
        #     ---------------------------------------------
        # 0   | Header                         | Map sel. |
        #     |-------------------------------------------|
        # 1   | Params |                                  |
        # 2   |        |                                  |
        # 3   |        |                                  |
        # 4   |        |               Map                |
        # 5   |        |                                  |
        #     |--------|                                  |
        # 6   |  Run   |                                  |
        #     |-------------------------------------------|
        # 7   |   Status info   |      Figure/output      |
        # 8   |                 |                         |
        # 9   |                 |                         |
        # 10  |                 |                         |
        # 11  ---------------------------------------------

        # Create the layout #[rowspan, colspan]
        grid = GridspecLayout(12, 10, height="1500px", width="auto")

        # Header and controls
        grid[0, :8] = self.header
        grid[0, 8:] = map_selection
        grid[1:6, 0:2] = parameter_selection
        grid[6, 0:2] = run_button

        # Status info, map and plot
        grid[1:7, 2:] = self.m  # map
        grid[7:, 0:4] = self.status_info
        grid[7:, 4:] = self.output_plot

        # Display using HBox children attribute
        self.children = [grid]

    ######################################
    # DEFINITION OF ALL UPDATE FUNCTIONS #
    ######################################

    # Update date
    def update_start_date(self, change):
        self.start_date = str(change.new)

        # Clear data load params to trigger data re-load
        self.timeseries_ds = None
        self.load_params = None
        self.query_params = None

    # Update date
    def update_end_date(self, change):
        self.end_date = str(change.new)

        # Clear data load params to trigger data re-load
        self.timeseries_ds = None
        self.load_params = None
        self.query_params = None

    # Update colour stretch
    def update_slider_percentile(self, change):
        self.vmin, self.vmax = change.new

    # Update power transform
    def update_slider_power(self, change):
        self.power = change.new

    # Update good data slider
    def update_floatslider_max_cloud_cover(self, change):
        self.max_cloud_cover = change.new

        # Clear data load params to trigger data re-load
        self.timeseries_ds = None
        self.load_params = None
        self.query_params = None

    # Enable unsharp masking and show/hide custom params
    def update_checkbox_unsharp_mask(self, change):
        self.unsharp_mask = change.new

        # Show unsharp masking params in menu if activated
        if change.new:
            self.text_unsharp_mask_radius.layout.display = "block"
            self.text_unsharp_mask_amount.layout.display = "block"
        else:
            self.text_unsharp_mask_radius.layout.display = "none"
            self.text_unsharp_mask_amount.layout.display = "none"

    # Change unsharp masking radius
    def update_text_unsharp_mask_radius(self, change):
        self.unsharp_mask_radius = change.new

    # Change unsharp masking amount
    def update_text_unsharp_mask_amount(self, change):
        self.unsharp_mask_amount = change.new

    # Enable rolling median and show/hide custom params
    def update_checkbox_rolling_median(self, change):
        self.rolling_median = change.new

        # Show rolling median params in menu if activated
        if change.new:
            self.text_rolling_median_window.layout.display = "block"
        else:
            self.text_rolling_median_window.layout.display = "none"

    # Change rolling median window
    def update_text_rolling_median_window(self, change):
        self.rolling_median_window = change.new

    # Override max size limit
    def update_checkbox_max_size(self, change):
        self.max_size = change.new

    # Add DE Africa Coastlines overlay
    def update_deacoastlines(self, change):
        self.deacoastlines = change.new

    # Apply cloud mask in load_ard
    def update_checkbox_cloud_mask(self, change):
        self.cloud_mask = change.new

        # Clear data load params to trigger data re-load
        self.timeseries_ds = None
        self.load_params = None
        self.query_params = None

    # Override min width
    def update_width(self, change):
        self.width = change.new

    # Override interval
    def update_interval(self, change):
        self.interval = change.new

    # Update resolution
    def update_text_resolution(self, change):
        self.resolution = change.new

        # Clear data load params to trigger data re-load
        self.timeseries_ds = None
        self.load_params = None
        self.query_params = None

    # Change layers shown on the map
    def update_dealayer(self, change):
        self.dealayer = change.new

        if change.new == "Landsat":
            self.text_resolution.value = 30

        else:
            self.text_resolution.value = 10

    # Update basemap
    def update_basemap(self, change):
        self.basemap = change.new
        update_map_layers(self)

    # Set imagery style
    def update_styles(self, change):
        self.style = change.new

        # Clear data load params to trigger data re-load
        self.timeseries_ds = None
        self.load_params = None
        self.query_params = None

    # Set output file format
    def update_output(self, change):
        self.output_format = change.new

    # Set output file format
    def update_dropdown_resampling(self, change):
        self.resample_freq = change.new

    def run_app(self, change):

        # Clear progress bar and output areas before running
        self.status_info.clear_output()
        self.output_plot.clear_output()

        # Verify that polygon was drawn
        if self.gdf_drawn is not None:

            with self.status_info:

                # Load data and add to attribute
                if self.timeseries_ds is None:
                    self.timeseries_ds = extract_data(self)

                else:
                    print("Using previously loaded data")

            if self.timeseries_ds is not None:

                with self.status_info:

                    # Create unique file name
                    centre_coords = self.gdf_drawn.geometry[0].centroid.coords[0][::-1]
                    site = reverse_geocode(coords=centre_coords)
                    fname = (
                        f"{self.dealayer}_{site}_{self.start_date}_"
                        f"{self.end_date}_{self.style}_{self.resolution:.0f}m."
                        f"{self.output_format}".replace(" ", "").replace(",", "").lower()
                    )

                    print(f"\nExporting animation for {site}.\nThis may take several minutes...")

                ############
                # Plotting #
                ############

                with self.output_plot:
                    plot_data(self, fname)

            else:
                with self.status_info:
                    print(
                        "No satellite data found in the selected area. "
                        "Please select a new rectangle over an area with "
                        "satellite imagery."
                    )

        else:
            with self.status_info:
                print('Please draw a valid rectangle on the map, then press "Generate animation".')
